;; Generated package description from tree-inspector.el  -*- no-byte-compile: t -*-
(define-package "tree-inspector" "0.4.0.20240322.113138" "Inspector tool for Emacs Lisp object that uses a treeview" '((emacs "27.1") (treeview "1.1.0")) :commit "da1e5a9df92bdbbaef439e34466760f6a5d707fb" :authors '(("Mariano Montone" . "marianomontone@gmail.com")) :maintainer '("Mariano Montone" . "marianomontone@gmail.com") :keywords '("debugging" "tool" "lisp" "development") :url "https://github.com/mmontone/emacs-tree-inspector")
